{-# OPTIONS -fglasgow-exts #-}
{-| Closed term with no meta variables can be compiled to something a bit more
    efficient.
-}
module Compilers.Run where

import Control.Monad.Trans
import qualified Data.Map as Map

import Syntax.Values
import Syntax.Pretty
import Syntax.Position
import Syntax.Name
import qualified Syntax.Internal as I

import TypeChecker.Abstractions
import TypeChecker.Monad
import TypeChecker.Monad.Context
import TypeChecker.Monad.Definitions
import TypeChecker.Monad.Errors
import TypeChecker.Monad.Options
import TypeChecker.Errors
import TypeChecker.Values.Refresh
import TypeChecker.Patterns

import Utils.Monads
import Utils.Pretty
import Utils.Error
import Utils.Sugar

-- The runtime language ---------------------------------------------------

type Map = Map.Map

data Val = Var Int [Val]
	 | Con Name [Val]
	 | Def Name [Val]	       -- ^ Only for things that have a definition. Always fully applied.
	 | Proj Val Prj [Val]
	 | Lam Val
	 | Pair Val Val
	 | Struct (Map String Val)
	 | Star			       -- ^ We ignore types
	 | Top
	 | Prim String Int ([NF] -> NF) [Val] -- ^ Primitive function with arity. Fully applied.

type Def	= ClauseTree
data ClauseTree = Done Val
		| Case (Map Name (Int,ClauseTree)) -- ^ Also binds the whole term
		| Split ClauseTree		 -- ^ Pairs (also binds the whole term)
		| Bind ClauseTree
    deriving Show

data Prj = Fst | Snd | Fld String
    deriving (Eq, Show)

instance Raiseable Val where
    raiseByFrom k n v =
	case v of
	    Var m vs	  -> Var (unDeBr $ rbf $ DeBr m) $ rbf vs
	    Con c vs	  -> Con c $ rbf vs
	    Def x vs	  -> Def x $ rbf vs
	    Proj v p vs   -> Proj (rbf v) p (rbf vs)
	    Lam v	  -> Lam $ raiseByFrom k (n + 1) v
	    Pair v w	  -> Pair (rbf v) (rbf w)
	    Struct xs	  -> Struct $ Map.map rbf xs
	    Star	  -> Star
	    Top		  -> Top
	    Prim x a f vs -> Prim x a f $ rbf vs
	where
	    rbf x = raiseByFrom k n x

instance Raiseable ClauseTree where
    raiseByFrom k n t =
	case t of
	    Done v  -> Done $ raiseByFrom k n v
	    Case m  -> Case $ Map.map rbfCase m
	    Split t -> Split $ raiseByFrom k (n + 1) t
	    Bind t  -> Bind $ raiseByFrom k (n + 1) t
	where
	    rbfCase (ar,t) = (ar, raiseByFrom k (n + 1) t)

instance Show Val where show = show . pretty
instance Show NF  where show = show . pretty

instance Pretty Val where
    prettyPrec pr v =
	case v of
	    _	| NatSugar n   <- sugar -> pretty n
		| ListSugar xs <- sugar -> brackets $ fsep $ punctuate comma $ map pretty xs
		where
		    sugar = sugaredView valSugar v
	    Var n vs	-> (text "@" <> pretty n) `apply` vs
	    Con c vs	-> prettyName c `apply` vs
	    Def f vs	-> prettyName f `apply` vs
	    Proj v p vs	-> (prettyPrec 2 v <> text "." <> pretty p) `apply` vs
	    Lam v	-> mparens (pr > 0) $ text "lam" <+> pretty v
	    Pair v w	-> cat [ text "("
			       , pretty v
			       , comma <+> pretty w
			       , text ")"
			       ]
	    Struct xs	-> sep $ text "struct"
			       : map pr (Map.toList xs)
		where
		    pr (x,v) = nest 2 $ text x <+> text "=" <+> pretty v

	    Star	-> text "Type"
	    Top		-> text "Top"
	    Prim x _ _ vs   -> (text "(prim)" <> text x) `apply` vs
	where
	    apply d [] = d
	    apply d vs = mparens (pr > 0) $ d <+> fsep (map (prettyPrec 1) vs)

	    valSugar = Sugar isZero isSuc isNil isCons
		where
		    isZero (Con (Name _ "zero" _) []  ) = True
		    isZero _				= False
		    isSuc  (Con (Name _ "suc"  _) [n] ) = Just n
		    isSuc _				= Nothing
		    isNil  (Con (Name _ "nil"  _) [_] ) = True
		    isNil _				= False
		    isCons (Con (Name _ "::"   _) vs  )
			| xs:x:_ <- reverse vs		= Just (x,xs)
		    isCons _				= Nothing

instance Pretty NF where
    prettyPrec pr v =
	case v of
	    _	| NatSugar n   <- sugar -> pretty n
		| ListSugar xs <- sugar -> brackets $ fsep $ punctuate comma $ map pretty xs
		where
		    sugar = sugaredView nfSugar v
	    NCon c vs	-> prettyName c `apply` vs
	    NLam v env	-> parens (text "lam" <+> pretty v)
			    <> braces (fsep $ punctuate comma $ map pretty env)
	    NPair v w	-> cat [ text "("
			       , pretty v
			       , comma <+> pretty w
			       , text ")"
			       ]
	    NStruct xs	-> sep $ text "struct"
			       : map pr (Map.toList xs)
		where
		    pr (x,v) = nest 2 $ text x <+> text "=" <+> pretty v
	    NStar	-> text "Type"
	    NTop	-> text "Top"
	where
	    apply d [] = d
	    apply d vs = mparens (pr > 0) $ d <+> fsep (map (prettyPrec 1) vs)

	    nfSugar = Sugar isZero isSuc isNil isCons
		where
		    isZero (NCon (Name _ "zero" _) []  ) = True
		    isZero _				 = False
		    isSuc  (NCon (Name _ "suc"	_) [n] ) = Just n
		    isSuc _				 = Nothing
		    isNil  (NCon (Name _ "nil"	_) [_] ) = True
		    isNil _				 = False
		    isCons (NCon (Name _ "::"	_) vs  )
			| xs:x:_ <- reverse vs		 = Just (x,xs)
		    isCons _				 = Nothing

instance Pretty Prj where
    pretty Fst	   = text "1"
    pretty Snd	   = text "2"
    pretty (Fld x) = text x

instance Eq Val where
    Var n vs	== Var m ws    = (n,vs)   == (m,ws)
    Con x vs	== Con y ws    = (x,vs)   == (y,ws)
    Def x vs	== Def y ws    = (x,vs)   == (y,ws)
    Proj v p vs == Proj w q ws = (v,p,vs) == (w,q,ws)
    Lam v	== Lam w       = v	  == w
    Pair v w	== Pair v' w'  = (v,w)	  == (v',w')
    Struct m	== Struct m'   = m	  == m'
    Star	== Star	       = True
    Top		== Top	       = True
    _		== _	       = False

-- Compiler ---------------------------------------------------------------

type Primitives = Map Name (Int, [NF] -> NF)

compileValue :: Primitives -> Value -> TC Val
compileValue prim v =
    sayWhen (fsep $ pwords "compiling" ++ [pretty v]) $
    sayWhat (fsep $ pwords "something went bad") $
    compile prim 0 =<< refresh v

-- | The first argument is the number of free variables.
compile :: Primitives -> Int -> Value -> TC Val
compile prim fv v =
    (verbose 12 $ liftIO $ putStrLn $ unwords ["compile", show fv, show v]) >>
    case vAppView v of
	VAVBId n vs
	    | n < fv	-> Var n <$> mapM comp vs
	    | otherwise -> failure $ notAClosedTerm v
	VAVId x _	-> failure $ namedVariable x
	VAVCon c _ _ vs -> Con c <$> mapM comp vs
	VAVDef f _ vs	->
	    case Map.lookup f prim of
		Nothing ->
		    do  DefVal _ ar' fv' def _ <- lookupDef f
			let ar = ar' + fv'
			case getClausesRT def of
			    []		    -> Con f <$> mapM comp vs
			    cs | nargs < ar -> comp $ eta (ar - nargs) v
			       | otherwise  -> Def f <$> mapM comp vs
		Just (ar,fun)
		    | nargs < ar    -> comp $ eta (ar - nargs) v
		    | otherwise	    -> Prim (nameString f) ar fun <$> mapM comp vs
	    where
		nargs = length vs
	VAVMeta _ _	-> failure $ metaVariable
	VAVProj v p vs	-> Proj <$> comp v <*> return (proj p) <*> mapM comp vs
	    where
		proj VFst	  = Fst
		proj VSnd	  = Snd
		proj (VField _ x) = Fld x
	VAVPair v w	-> Pair <$> comp v <*> comp w
	VAVLam f	-> Lam <$> compile prim (fv + 1) (absBody f)
	VAVStruct xs	-> Struct . Map.fromList <$> mapM fld xs
	    where fld (x,(_,v)) = (,) x <$> comp v
	VAVTop		-> return Top
	VAVSort _	-> return Star
	VAVPi _ _	-> return Star
	VAVSigma _ _	-> return Star
    where
	comp = compile prim fv

eta :: Int -> Value -> Value
eta 0 v = v
eta n v = vLam $ mkAbs "x" $ eta (n - 1) (raise v `vApp` vBId 0)

compileDefs :: Primitives -> TC (Map Name Def)
compileDefs prim =
    do	defs <- Map.toList <$> getDefinitions
	ds   <- concat <$> mapM comp defs
	return $ Map.fromList ds
    where
	comp (x,d@(DefVal _ _ _ def _)) =
	    case cs of
		[]  -> return []
		_   -> (\d -> [(x,d)]) <$> compileDef prim x d
	    where
		cs = getClausesRT def

compileDef :: Primitives -> Name -> DefValue -> TC Def
compileDef prim x (DefVal _ _ _ def _) =
    sayWhen (fsep $ pwords $ "compiling definition of " ++ show x) $
    sayWhat (fsep $ pwords "something went bad") $
    do	def <- mergeClauseTrees <$> mapM (compileClause prim) cl
	verbose 10 $ liftIO $ putStrLn $ "compiled " ++ show x ++ ": " ++ show def
	return def
    where
	cl = getClausesRT def

compileClause :: Primitives -> ClauseValue -> TC ClauseTree
compileClause prim (ClauseVal ps w) = compilePatterns prim n n (unLam n w) ps
    where
	n = length $ concatMap patternVars ps

-- | Strip off a number of lambdas.
unLam :: Int -> Value -> Value
unLam 0 v = v
unLam n v =
    case vView v of
	VLam f	-> unLam (n - 1) $ absBody f
	_	-> error $ badClause n v

-- | @compilePatterns n k v ps@. @n = @ /free variables in v/, @k = @ /bound
--   variables ps/.
compilePatterns :: Primitives -> Int -> Int -> Value -> [I.Pattern] -> TC ClauseTree
compilePatterns prim n 0 w []     = Done <$> compile prim n w
compilePatterns prim n k w []     = error $ "impossible: Compilers.Run.compilePatterns " ++ show (n,k,w,[]::[Int])
compilePatterns prim n k w (p:ps) =
    case p of
	I.VarP x	  -> Bind <$> compilePatterns prim n (k - 1) w ps
	I.WildP _	  -> Bind <$> compilePatterns prim (n + 1) k (shift w) ps
	I.ConP _ c hqs qs -> Case . Map.singleton c . (,) ar
				<$> compilePatterns prim (n + 1) k (shift w) (hqs ++ qs ++ ps)
	    where
		ar = length (hqs ++ qs)
	I.PairP _ p q	  -> Split <$> compilePatterns prim (n + 1) k (shift w) (p:q:ps)
    where
	shift = raiseByFrom 1 k

mergeClauseTrees :: [ClauseTree] -> ClauseTree
mergeClauseTrees ts = foldr1 merge ts
    where
	merge (Done v)	 (Done _)   = Done v
	merge (Bind t1)  (Bind t2)  = Bind $ merge t1 t2
	merge (Split t1) (Bind t2)  = Split $ merge t1 (bind 2 t2)
	merge (Bind t1)  (Split t2) = Split $ merge (bind 2 t1) t2
	merge (Split t1) (Split t2) = Split $ merge t1 t2
	merge (Case m)	 (Bind t)   = Case $ Map.map (flip mL t) m
	merge (Bind t)   (Case m)   = Case $ Map.map (mR t) m
	merge (Case m1)	 (Case m2)  = Case $ Map.unionWith mLR m1 m2
	merge (Done _)	  _	    = error "impossible: Compilers.Run.merge (Done _) _"
	merge  _	 (Done _)   = error "impossible: Compilers.Run.merge _ (Done _)"
	merge (Case _)	 (Split _)  = error "impossible: Compilers.Run.merge (Case _) (Split _)"
	merge (Split _)	 (Case _)   = error "impossible: Compilers.Run.merge (Split _) (Case _)"

	mL (ar,t1) t2	= (ar, merge t1 $ bind ar t2)
	mR t1 (ar,t2)	= (ar, merge (bind ar t1) t2)
	mLR (ar,t1) (ar',t2)
	    | ar /= ar' = error "impossible: Compilers.Run.mkLR"
	    | otherwise = (ar, merge t1 t2)

	bind 0 t = t
	bind n t = Bind $ raiseByFrom 1 1 $ bind (n - 1) t

-- Compiler errors --------------------------------------------------------

notAClosedTerm v = customError $ "not a closed term: " ++ show v
namedVariable x  = customError $ "unexpected named variable " ++ show x
metaVariable	 = customError "unexpected metavariable"
badClause n v	 = "expected " ++ show n ++ " more lambdas in " ++ show v

-- Run --------------------------------------------------------------------

type Env = [NF]
data NF  = NCon Name [NF]
	 | NLam Val Env
	 | NPair NF NF
	 | NStruct (Map String NF)
	 | NStar
	 | NTop
    deriving (Eq)

-- | Call by value.
evaluate :: Map Name Def -> Val -> NF
evaluate defs v = eval defs [] v

eval defs env v =
    case v of
	Var n vs      -> (env !! n) `app` evals vs
	Con c vs      -> NCon c $ evals vs
	Def f vs      -> match (look f) $ evals vs
	Proj v p vs   -> proj (eval defs env v) p `app` evals vs
	Pair v w      -> NPair (eval defs env v) (eval defs env w)
	Struct m      -> NStruct $ Map.map (eval defs env) m
	Star	      -> NStar
	Top	      -> NTop
	Lam v	      -> NLam v env
	Prim _ a f vs -> f ws0 `app` ws1
	    where (ws0,ws1) = splitAt a $ evals vs
    where
	look f = case Map.lookup f defs of
		    Nothing -> error $ "impossible: Compilers.Run.look " ++ show f
		    Just d  -> d
	lookField m x =
	    case Map.lookup x m of
		Nothing -> error $ "impossible: Compilers.Run.lookField " ++ show x
		Just v	-> v

	branch m x =
	    case Map.lookup x m of
		Nothing	    -> error $ "Pattern match failure: No case for " ++ show x ++ " in " ++ show m
		Just (_,t)  -> t

	evals = map (eval defs env)

	proj (NPair v w)  Fst	 = v
	proj (NPair v w)  Snd	 = w
	proj (NStruct m) (Fld x) = lookField m x
	proj _		 _	 = error $ "impossible Compilers.Run.proj"

	app = nfApply defs

	match t vs = mtch [] t vs
	    where
		mtch env (Done v)   vs			= eval defs env v `app` vs
		mtch env (Bind t)  (v:vs)		= mtch (v:env) t vs
		mtch env (Split t) (v@(NPair v1 v2):vs) = mtch (v:env) t (v1:v2:vs)
		mtch env (Case m)  (v@(NCon c ws):vs)	=
		    mtch (v:env) (branch m c) (ws ++ vs)
		mtch env t vs	= error $ "impossible Compilers.Run.match (" ++ show t ++ ") " ++ show vs

nfApply :: Map Name Def -> NF -> [NF] -> NF
nfApply defs  v		  []	 = v
nfApply defs (NLam v env) (w:ws) = nfApply defs (eval defs (w:env) v) ws
nfApply defs (NCon c vs)   ws	 = NCon c $ vs ++ ws
nfApply defs  NTop	   ws	 = NTop
nfApply defs  v		   ws	 = error $ "impossible Compilers.Run.nfApply (" ++ show v ++ ") " ++ show ws

